@@ -2,7 +2,7 @@
  * GNOME panel utils
  * (C) 1997, 1998, 1999, 2000 The Free Software Foundation
  * Copyright 2000 Helix Code, Inc.
- * Copyright 2000 Eazel, Inc.
+ * Copyright 2000,2001 Eazel, Inc.
  *
  * Authors: George Lebl
  *          Jacob Berkman
@@ -17,6 +17,8 @@
 
 #include "panel-include.h"
 
+#include "icon-entry-hack.h"
+
 extern GlobalConfig global_config;
 
 extern GSList *applets;
@@ -30,6 +32,82 @@ panel_show_help (const char *path)
 	gnome_help_display (NULL, &help_entry);
 }
 
+static char *
+panel_gnome_help_path (const char *docpath)
+{
+	char *fullpath, *app, *p, *path, *uri;
+
+	app = g_strdup (docpath);
+
+	p = strchr (app, '/');
+
+	if (p == NULL) {
+		g_free (app);
+		return NULL;
+	}
+
+	path = p+1;
+	*p = '\0';
+
+	fullpath = gnome_help_file_path (app, path);
+
+	g_free (app);
+
+	if ( ! panel_file_exists (fullpath)) {
+		g_free (fullpath);
+		fullpath = NULL;
+	}
+
+	uri = g_strconcat ("ghelp:", fullpath, NULL);
+	g_free (fullpath);
+
+	return uri;
+}
+
+static char *
+panel_kde_help_path (const char *docpath)
+{
+	GList *li;
+
+	if ( ! panel_file_exists (KDE_DOCDIR))
+		return NULL;
+
+	for (li = gnome_i18n_get_language_list ("LC_MESSAGES");
+	     li != NULL;
+	     li = li->next) {
+		char *fullpath = g_strdup_printf ("%s/HTML/%s/%s",
+						  KDE_DOCDIR,
+						  li->data,
+						  docpath);
+		if (panel_file_exists (fullpath)) {
+			char *uri = g_strconcat ("ghelp:", fullpath, NULL);
+			g_free (fullpath);
+			return uri;
+		}
+		g_free (fullpath);
+	}
+	return NULL;
+}
+
+char *
+panel_gnome_kde_help_path (const char *docpath)
+{
+	char *path;
+
+	if (string_empty (docpath))
+		return NULL;
+
+	if (panel_is_url (docpath))
+		return g_strdup (docpath);
+
+	path = panel_gnome_help_path (docpath);
+
+	if (path == NULL)
+		path = panel_kde_help_path (docpath);
+
+	return path;
+}
+
 gboolean
 string_is_in_list(const GSList *list, const char *text)
 {
@@ -112,6 +190,7 @@ create_icon_entry(GtkWidget *table,
 	gtk_widget_show(wlabel);
 
 	entry = gnome_icon_entry_new(history_id,_("Browse"));
+	hack_icon_entry (GNOME_ICON_ENTRY (entry));
 	gnome_icon_entry_set_pixmap_subdir(GNOME_ICON_ENTRY(entry), subdir);
 	if (text)
 		gnome_icon_entry_set_icon(GNOME_ICON_ENTRY(entry),text);
@@ -872,3 +951,21 @@ get_requisition_height (GtkWidget *widget)
 
 	return req.height;
 }
+
+/* is url showable by gnome_url_show */
+gboolean
+panel_is_url (const char *url)
+{
+	if (strncmp (url, "http://", strlen ("http://")) == 0 ||
+	    strncmp (url, "https://", strlen ("https://")) == 0 ||
+	    strncmp (url, "gopher://", strlen ("gopher://")) == 0 ||
+	    strncmp (url, "ftp://", strlen ("ftp://")) == 0 ||
+	    strncmp (url, "file:", strlen ("file:")) == 0 ||
+	    strncmp (url, "ghelp:", strlen ("ghelp:")) == 0 ||
+	    strncmp (url, "help:", strlen ("help:")) == 0 ||
+	    strncmp (url, "man:", strlen ("man:")) == 0 ||
+	    strncmp (url, "info:", strlen ("info:")) == 0)
+		return TRUE;
+	else
+		return FALSE;
+}
